/**
 * 
 */
package sample.auction.behaviours;

import java.util.LinkedList;
import java.util.List;

import jade.core.AID;
import jade.core.Agent;
import jade.core.behaviours.Behaviour;
import jade.domain.DFService;
import jade.domain.FIPAException;
import jade.domain.FIPAAgentManagement.DFAgentDescription;
import jade.domain.FIPAAgentManagement.ServiceDescription;
import jade.lang.acl.ACLMessage;

/**
 * I[NVjAG[WFgBehavior
 */
public class AuctioneerBehaviour extends Behaviour {

	private static final long serialVersionUID = 8233468534866591311L;

	/** G[WFg̓̏Ԃ̒` */
	private static final int INITIAL  = 0;
	private static final int CFP      = 1;
	private static final int RECV_BID = 2;
	private static final int FINISH   = 3;

	/** G[WFg̓̏ */
	private int state_ = INITIAL;
	/** G[WFg̓̏ItO */
	private boolean done_ = false;
	/** G[WFg̓ɂāAG[ǂ̃tO */
	private boolean fail_ = false;

	/** rb_[̃Xg */
	private List<DFAgentDescription> dfdList_ = null;
	/** ݂CFPɑ΂rbh̃Xg */
	private List<ACLMessage> bidList_ = new LinkedList<ACLMessage>();
	/** ݂̍ő̃rbh̉i */
	private int maxBid_ = 0;
	/** ŒᗎDi */
	private int criteria_ = 0;
	/** ݂̍ōD */
	private AID winner_ = null;
	
	/**
	 * RXgN^
	 * @param agent
	 */
	public AuctioneerBehaviour(Agent agent) {
		super(agent);
	}

	/**
	 * Behaviour̓̒`
	 */
	public void action() {
		switch (state_) {
		case INITIAL:
			// rb_[̌ƃI[NV̊Jnʒm
			beginAuction();
			state_ = CFP;
			break;
		case CFP:
			// @rbh̕W
			sendCFP();
			state_ = RECV_BID;
			break;
		case RECV_BID:
			// rbh̎MƑI
			int result = recvBid();
			if (result == 0) {
				state_ = CFP;
			} else if (result > 0) {
				state_ = FINISH; 
			}
			break;
		case FINISH:
			// I[NV̏Iʒm
			endAuction();
			done_ = true;
			break;
		}
	}

	/**
	 * Is
	 * @return I[NVIA܂́AsĂΏIƔ肷
	 */
	public boolean done() {
		return done_ || fail_;
	}

	/**
	 * I[NV̊Jns
	 */
	private void beginAuction() {
		// I[NV̏ݒ
		// rbh̐ݒ
		maxBid_ = (int)(5000 * Math.random());
		// ]Di̐ݒ
		criteria_ = (int)(maxBid_ + 5000 * Math.random());
		
		System.out.println("Auctioneer::beginAuction()");
		System.out.println("  initial:  " + maxBid_);
		System.out.println("  criteria: " + criteria_);

		// rb_[̌
		dfdList_ = new LinkedList<DFAgentDescription>();
		try {
			ServiceDescription sd = new ServiceDescription();
			sd.setName("EnglishAuction");
			sd.setType("Bidder");
			DFAgentDescription dfd = new DFAgentDescription();
			dfd.addOntologies("ea-ontology");
			dfd.addServices(sd);
			DFAgentDescription[] list = DFService.search(myAgent, dfd);
			if (list == null) {
				fail_ = true;
				return;
			}
			for (int i = 0; i < list.length; ++i) {
				DFAgentDescription desc = (DFAgentDescription)list[i];
				dfdList_.add(desc);
			}
		} catch (FIPAException e) {
			e.printStackTrace();
			fail_ = true;
		}

		// rb_[ɒʒm
		ACLMessage msg = new ACLMessage(ACLMessage.INFORM);
		msg.setSender(myAgent.getAID());
		for (int i = 0; i < dfdList_.size(); ++i) {
			DFAgentDescription dfd = (DFAgentDescription)dfdList_.get(i);
			msg.addReceiver(dfd.getName());
		}
		msg.setContent("begin");
		myAgent.send(msg);
	}
	
	/**
	 * rb_[ɃI[NṼrbhWʒms
	 */
	private void sendCFP() {
		// rb_[ɒʒm
		ACLMessage msg = new ACLMessage(ACLMessage.CFP);
		msg.setSender(myAgent.getAID());
		for (int i = 0; i < dfdList_.size(); ++i) {
			DFAgentDescription dfd = (DFAgentDescription)dfdList_.get(i);
			msg.addReceiver(dfd.getName());
		}
		msg.setContent(String.valueOf(maxBid_));
		myAgent.send(msg);
		bidList_.clear();
		
		System.out.println("Auctioneer::sendCFP()");
		System.out.println("  max bid:  " + maxBid_);
	}
	
	/**
	 * rbhM
	 * @return rbhMAׂẴrb_[̃rbhMł0܂1ԂB܂M̃rb_[ꍇ-1ԂB
	 */
	private int recvBid() {
		ACLMessage msg = myAgent.receive();
		if (msg != null) {
			System.out.println("Auctioneer::recvBid()");
			System.out.println("  " + msg.getSender().getLocalName());
			System.out.println("  " + msg.getContent());
			
			bidList_.add(msg);
			if (bidList_.size() == dfdList_.size()) {
				return selectBid();
			} else {
				return -1;
			}
		} else {
			block();
			return -1;
		}
	}
	
	/**
	 * Mrbh󗝂rbhIoƂƂɁArb_[ɃX|XԂ
	 * @return ̃rbhɂI[NVIꍇ1AłȂꍇ0Ԃ
	 */
	private int selectBid() {
		int max = maxBid_;
		AID win = null;
		
		// ő̃rbhI
		for (int i = 0; i < bidList_.size(); ++i) {
			ACLMessage m = (ACLMessage)bidList_.get(i);
			int bid = Integer.parseInt(m.getContent());
			if (max < bid) {
				max = bid;
				win = m.getSender();
			}
		}
		
		// rb_[ɑ΂ăX|X𑗐M
		for (int i = 0; i < bidList_.size(); ++i) {
			ACLMessage m = (ACLMessage)bidList_.get(i);
			ACLMessage res = m.createReply();
			res.setSender(myAgent.getAID());
			if (win == m.getSender()) {
				res.setPerformative(ACLMessage.ACCEPT_PROPOSAL);
			} else {
				res.setPerformative(ACLMessage.REJECT_PROPOSAL);
			}
			myAgent.send(res);
		}
		
		maxBid_ = max;
		winner_ = win;
		
		// ŒᗎDirbhĂ΁AI[NVIƂ
		if (max < criteria_) {
			return 0;
		} else {
			return 1;
		}
	}
			
	/**
	 * I[NV̏Iʒm
	 */
	private void endAuction() {
		// rb_[ɒʒm
		ACLMessage msg = new ACLMessage(ACLMessage.INFORM);
		msg.setSender(myAgent.getAID());
		for (int i = 0; i < dfdList_.size(); ++i) {
			DFAgentDescription dfd = (DFAgentDescription)dfdList_.get(i);
			msg.addReceiver(dfd.getName());
		}
		msg.setContent("end");
		myAgent.send(msg);		

		System.out.println("Auctioneer::endAuction()");
		System.out.println("  winner: " + winner_.getLocalName());
		System.out.println("  price:  " + maxBid_);
	}

}
